/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.components.script;

import com.vii.brillien.core.component.SuperPresence;
import com.vii.brillien.core.management.BrillienServices;
import com.vii.brillien.ignition.transport.BrillienCommunication;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.kernel.annotations.PresenceService;
import com.vii.brillien.kernel.annotations.lifecycle.Resident;
import com.vii.streamline.annotation.P;
import com.vii.streamline.services.TypeServices;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.script.ScriptServices;
import com.vii.streamline.structures.collections.InnerMap;
import sun.org.mozilla.javascript.internal.Function;
import sun.org.mozilla.javascript.internal.Scriptable;

import javax.script.ScriptContext;
import javax.script.ScriptEngine;
import java.util.Map;

/**
 * Special Presence type representing a software-based Processor receiving behavior by scripts as parameter of a special received communication.
 */
@PresenceService( logLevel = PresenceService.FINE )
@Resident
public class Processor<R> extends SuperPresence<R> {

    protected ScriptEngine                          testEngine;
    protected ScriptContext                         testContext;

    protected ScriptEngine                          engine;
    protected ScriptContext                         context;

    protected Map<String, ScriptMessageProcessor>   scriptedAssignment;


    protected void initScripting(){
        log("Initiating scriptengine..." );

        testEngine  = ScriptServices.createScriptEngine();
        testContext = ScriptServices.createScriptContext( testEngine );

        engine  = ScriptServices.createScriptEngine();
        context = ScriptServices.createScriptContext( engine );

        // ToDo: is there any resource missing?
        engine.put( "presence", this );
        engine.put( "liaison", BrillienServices.liaison );
        engine.put( "apiKeyPresenceName", getApiKeyPresenceName() );
        engine.put( "ssoPresenceName", getSSOPresenceName() );
    }


    @Override
    public void plan() throws BrillienException {
        initScripting();

        scriptedAssignment = new InnerMap<String, ScriptMessageProcessor>();
    }

    protected void eval(ScriptEngine engine, ScriptContext context, String messageProcessor, String script) throws StreamLineException, BrillienException {
        ScriptServices.eval( engine, context, script );
        Object ff = context.getAttribute( messageProcessor );
        if( ff == null || ff == Scriptable.NOT_FOUND || !(ff instanceof Function) )
            throw new BrillienException(StreamLineException.ERROR_INVALID_PARAMS, "Script does not match to the presumed messageProcessor: " + messageProcessor);
    }

    public String publishMessageProcessor( @P(name = "messageProcessor") ScriptMessageProcessor messageProcessor ) throws BrillienException {
        if( messageProcessor == null || messageProcessor.getName() == null || messageProcessor.getScript() == null || messageProcessor.getParameters() == null )
            throw new BrillienException(StreamLineException.ERROR_INVALID_PARAMS, "Parameter \"messageProcessor\" need to be defined properly.", messageProcessor);

        String name = messageProcessor.getName();
        try {
            log("Mapping types:" + messageProcessor.getParameters() );

            Map<String, Class> types = new InnerMap<String, Class>();
            for (String parameterName : messageProcessor.getParameters().keySet()) {
                String parameterType = messageProcessor.getParameters().get( parameterName );

                types.put( parameterName, TypeServices.getJsonType( parameterType ));
            }

            log("Evaluating script:" + messageProcessor.getScript() );

            eval( testEngine, testContext, name, messageProcessor.getScript() );
            eval(engine, context, name, messageProcessor.getScript());

            getPresenceManager().addPresenceService( name, types );
            BrillienServices.addPublisher(this.getName(), name);
            scriptedAssignment.put(name, messageProcessor);
        } catch (StreamLineException e) {
            throw new BrillienException( e.getErrorCode(), e.getMessage(), e.getPayLoad() );
        } catch (BrillienException e) {
            throw e;
        }

        return BrillienServices.DEFAULT_ANSWER;
    }

    public String removeMessageProcessor( @P(name = "messageProcessor") String messageProcessor ) throws BrillienException {
        scriptedAssignment.remove( messageProcessor );

        return BrillienServices.DEFAULT_ANSWER;
    }


    @Override
    protected Object invokeMessageProcessor( BrillienCommunication message ) throws BrillienException {
        if( context != null ){
            try{
                String msg = getSubject(message);
                if( !scriptedAssignment.containsKey(msg) )
                    throw new BrillienException(StreamLineException.ERROR_INVALID_PARAMS, "Could not find the requested service: " + msg);

                Object ff = context.getAttribute( msg );
                if( ff != null && ff != Scriptable.NOT_FOUND && (ff instanceof Function) ){
                    log("Invoking javascript code... " );
                    return ScriptServices.call( engine, context, msg, message.getParameters() );
                } else
                    throw new BrillienException(StreamLineException.ERROR_INTERNAL_ERROR, "Could not find the script code for this message: " + msg);
            } catch( Throwable t ){
                exceptionLog( Processor.class.getName(), "invokeMessageProcessor", t );
            }
        }

        return super.invokeMessageProcessor( message );
    }
    /*
    public static void main(String[] args) throws BrillienException {
        Processor p = new Processor();
        p.plan();

        p.publishMessageProcessor( new ScriptMessageProcessor(
                "hello", "function hello(){ print(\'Helloka!\');}"
        ) );

        BrillienCommunication comm = new BrillienCommunication();
        comm.setSubject( "hello" );
        comm.setParameters(new InnerMap<String, Object>() );
        p.invokeMessageProcessor( comm );
    }
    */
}

