/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */
package com.vii.brillien.services.rest;

import com.vii.brillien.core.management.BrillienServices;
import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.ignition.transport.BrillienCommunication;
import com.vii.brillien.ignition.transport.BrillienMediator;
import com.vii.brillien.ignition.transport.LoaderComplyer;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.kernel.axiom.sso.Session;
import com.vii.streamline.apikey.ApiKeyValidator;
import com.vii.streamline.rest.RestDispatcherServlet;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.reflection.Invocation;
import com.vii.streamline.sso.SessionValidator;
import com.vii.streamline.structures.collections.InnerMap;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;

/**
 * Specialised Servlet class representing a dispatcher providing a RESTful web services interface delegating all incoming calls to the internal business logic.
 */
public class BrillienRestServlet extends RestDispatcherServlet {

    /**
     * Singleton instance to be published by the webserver
     */
    protected static BrillienRestServlet    servletInstance;

    /**
     * mediator used to communicate with presences
     */
    protected BrillienMediator              mediator;

    static {
        servletInstance = new BrillienRestServlet( );
    }

    /**
     * Following the singleton model, retrieves the single instance created
     */
    public static BrillienRestServlet getServletInstance() {
        return servletInstance;
    }


    {
        initMediatorContext();
    }

    protected void initMediatorContext() {

    }

    protected void setupMediatorContext(ServletConfig servletConfig) {

    }

    protected BrillienMediator getMediator() throws ServletException {
        if( mediator == null ){
            String username = getServletConfig().getInitParameter("username");
            String password = getServletConfig().getInitParameter("password");
            if( username == null ){
                username = "dispatcher" + this.getClass().getSimpleName(); password = "dispatcher";
            }
            try {
                if( mediator == null ){
                    setupMediatorContext( getServletConfig() );
                    mediator = BrillienContext.newMediator(username, password, new LoaderComplyer() {
                        @Override
                        public ClassLoader getLoader(String reference) {
                            try {
                                return BrillienServices.getPresenceManagerByName(
                                        BrillienServices.liaison,reference).getClassLoader();
                            } catch (BrillienException e) {
                                e.printStackTrace();
                                return this.getClass().getClassLoader();
                            }
                        }
                    });
                    mediator.connect();
                }
            } catch (BrillienException e) {
                throw new ServletException( e.getMessage(), e );
            }
        }
        return mediator;
    }

    @Override
    protected boolean checkApiKeyRequired(){
        boolean apiKeyRequired = super.checkApiKeyRequired();

        return  BrillienServices.liaison != null
                ? BrillienServices.liaison.getApiKeyPresenceName() != null
                : apiKeyRequired;
    }

    @Override
    protected boolean checkSsoRequired(){
        boolean ssoRequired = super.checkSsoRequired();

        return  BrillienServices.liaison != null
                ? BrillienServices.liaison.getSSOPresenceName() != null
                : ssoRequired;
    }

    @Override
    protected ApiKeyValidator getApiKeyValidator() throws ServletException {
        try {
            return (ApiKeyValidator)BrillienServices.getPresence( BrillienServices.liaison, BrillienServices.liaison.getApiKeyPresenceName() );
        } catch (BrillienException e) {
            throw new ServletException( e );
        }
    }
    @Override
    protected SessionValidator getSessionValidator() throws ServletException {
        try {
            return (SessionValidator)BrillienServices.getPresence( BrillienServices.liaison, BrillienServices.liaison.getSSOPresenceName() );
        } catch (BrillienException e) {
            throw new ServletException( e );
        }
    }

    @Override
    public void init() throws ServletException {
        super.init();

        getMediator();
    }

    public boolean hasMessageProcessorReturnValue( Invocation invocation ) throws StreamLineException{
        try{
            Boolean res = mediator.sendGet(
                    Session.createSession( invocation.getSessionId() ),
                    invocation.getEntity(), invocation.getApiVersion(),
                    "hasMessageProcessorReturnValue",
                    new InnerMap<String, Object>( "messageProcessorName", invocation.getMethod() )
            ).getResponse();
            return res;
        } catch (BrillienException e) {
            throw new StreamLineException( e );
        }
    }

    public boolean isMethodReckognised(Invocation invocation) {
        return BrillienServices.isServiceExists(invocation);
    }

    protected void error(Exception e){
        if( logger != null )
            logger.error( e.getMessage(), e );
        else
            e.printStackTrace();    
    }
    
    @Override
    protected Object call(Invocation invocation) throws StreamLineException {
        if( !isMethodReckognised( invocation ) )
            throw new StreamLineException(StreamLineException.ERROR_INVALID_REQUEST, "Invocation's target cannot be identified.", (invocation.getEntity() == null ? "" : invocation.getEntity() + ".") + invocation.getMethod() );

        Session session = ssoRequired ? Session.createSession( invocation.getSessionId() ) : null;

        try {
            if( logger != null )
                logger.debug(
                    "Calling Presence " +
                    invocation.getEntity()
                );
            boolean hasReturnValue = hasMessageProcessorReturnValue( invocation );
            Object payload = null;
            if( hasReturnValue ){
                BrillienCommunication c = mediator.sendGet(
                        session, BrillienContext.RECONNECTION_TIMEOUT,
                        invocation.getEntity(),invocation.getApiVersion(),
                        invocation.getMethod(),
                        invocation.getParameters()
                );
                payload = c.getResponse();
            } else {
                mediator.sendSet(
                        session,
                        invocation.getEntity(),invocation.getApiVersion(),
                        invocation.getMethod(),
                        invocation.getParameters()
                );
                payload = "Success.";
            }

            if( logger != null )
                logger.debug(
                    "Response received: " + payload
                );

            return payload;
        } catch (BrillienException e) {
            error( e );
            throw new StreamLineException( e.getErrorCode(), e.getMessage(), e.getValue() );
        } catch (StreamLineException e) {
            error( e );
            throw e;
        } catch (Exception e) {
            error( e );
            throw new StreamLineException( e.getMessage(), e );
        }
    }

    @Override
    public void destroy() {
        try {
            if (mediator != null)
                mediator.disconnect();
        } catch (BrillienException e) {
        }
        BrillienContext.cleanup();
    }

}
