/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.web;

import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.kernel.BrillienException;
import com.vii.streamline.ZeroServlet;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.structures.collections.InnerList;
import com.vii.streamline.web.HttpServices;
import org.glassfish.grizzly.comet.CometContext;
import org.glassfish.grizzly.comet.CometEngine;
import org.glassfish.grizzly.comet.CometEvent;
import org.glassfish.grizzly.comet.DefaultCometHandler;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;

/**
 * Comet service to catch log records published to listening clients
 */
public class LogCatcherServlet extends ZeroServlet {

    /**
     * Singleton instance to be published by the webserver
     */
    protected static LogCatcherServlet          servletInstance;

    protected static Level                      interstedLevel;

    protected ConcurrentLinkedQueue<LogRecord>  records;

    protected String                            contextPath;

    protected boolean                           ready;

    protected class LogHandler extends DefaultCometHandler<HttpServletResponse> {

        private HttpServletResponse httpResponse;

        public LogHandler(HttpServletResponse httpResponse) {
            this.httpResponse = httpResponse;
            //System.out.println("Created...");
        }

        protected void sendData() throws IOException {
            InnerList<LogRecord> recordsToSend = new InnerList<LogRecord>( records );
            records.clear();

            try {
                HttpServices.sendJSONMessage(null, httpResponse, recordsToSend);
            } catch (StreamLineException e) {
                sendError(null, httpResponse, StreamLineException.ERROR_INTERNAL_ERROR, e.getMessage(), null);
            }
        }

        @Override
        public void onEvent(CometEvent event) throws IOException {
            //System.out.println("on Event: "+ event);

            if (CometEvent.Type.NOTIFY == event.getType()) {
                sendData( );
                event.getCometContext().resumeCometHandler(this);
            }
        }

        @Override
        public void onInterrupt(CometEvent event) throws IOException {
            //System.out.println("Interrupt: "+ event);
            event.getCometContext().resumeCometHandler(this);
        }
    }

    static {
        servletInstance = new LogCatcherServlet( );
        interstedLevel  = Level.INFO;
    }

    /**
     * Following the singleton model, retrieves the single instance created
     */
    public static LogCatcherServlet getServletInstance() {
        return servletInstance;
    }

    @Override
    public void init() throws ServletException {
        super.init();

        ready = false;

        records = new ConcurrentLinkedQueue<LogRecord>();

        logger.info( this.getClass().getSimpleName() + " just joined to the log handlers...");

        frameBasedValidation = false;

        BrillienContext.mainLogger.addHandler( new Handler() {
            @Override
            public synchronized Level getLevel() {
                return interstedLevel;
            }

            @Override
            public void publish(LogRecord logRecord) {
                try {
                    LogCatcherServlet.this.publish( logRecord );
                } catch (StreamLineException e) { } catch (IOException e) { }
            }

            @Override
            public void flush() {
            }

            @Override
            public void close() throws SecurityException {
            }
        });
        BrillienContext.basicLog(Level.INFO, this.getClass().getSimpleName() + " is capturing logs...");
    }

    @Override
    public void init(ServletConfig config) throws ServletException {
        super.init(config);

        contextPath = config.getServletContext().getContextPath() + this.getServletName();

        CometEngine engine = CometEngine.getEngine();
        CometContext cometContext = engine.register( contextPath );
        cometContext.setExpirationDelay(5 * 30 * 1000);

        ready = true;
    }

    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp) throws javax.servlet.ServletException, java.io.IOException  {
        if( !frameBasedValidation || checkCallPreconditions(req, resp) ){
            CometEngine engine = CometEngine.getEngine();
            CometContext<HttpServletResponse> context = engine.getCometContext(contextPath);
            context.addCometHandler(new LogHandler(resp));
        }
    }


    @Override
    protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws javax.servlet.ServletException, java.io.IOException  {
        try {
            String requestString = HttpServices.getRequestBody(req);

            logger.info("Attempt to set log level:" + requestString);

            if( requestString != null && requestString.length()>0 )
                interstedLevel = Level.parse( requestString );
            else
                throw new BrillienException(StreamLineException.ERROR_INVALID_PARAMS, "Not proper posted value.");

            sendResponse(req, resp, "OK. Level: " + interstedLevel.getLocalizedName());
        } catch (StreamLineException e) {
            sendError( req, resp, e.getErrorCode(), e.getMessage(), e.getPayLoad() );
        } catch (BrillienException e) {
            sendError( req, resp, e.getErrorCode(), e.getMessage(), e.getValue() );
        }
    }

    public void publish(LogRecord record) throws StreamLineException, IOException {
        while( records.size() > BrillienContext.LOGVIEWER_RECORD_CACHE )
            records.poll();

        if( record.getLevel().intValue() >= interstedLevel.intValue() || record.getThrown() != null ){
            records.offer( record );


            if( ready ){
                CometContext<HttpServletResponse> context = CometEngine.getEngine().getCometContext(contextPath);
                context.notify(null);
            }
        }
    }

}
