/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.web.purity;

import com.vii.brillien.core.management.BrillienServices;
import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.ignition.transport.BrillienCommunication;
import com.vii.brillien.ignition.transport.BrillienMediator;
import com.vii.brillien.ignition.transport.LoaderComplyer;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.kernel.axiom.transport.Communication;
import com.vii.streamline.apikey.ApiKeyValidator;
import com.vii.streamline.json.JsonRpcError;
import com.vii.streamline.purity.PureServlet;
import com.vii.streamline.services.RequestStatics;
import com.vii.streamline.services.ThreadServices;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.json.JsonServices;
import com.vii.streamline.sso.SessionValidator;
import com.vii.streamline.web.HttpServices;
import org.svenson.SvensonRuntimeException;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.HashMap;

/**
 * Specialized HttpServlet to manage so called Pure-protocol, meaning to delegate BrillienCommunication calls to presences.
 */
public class PurityServlet extends PureServlet {

    /**
     * Singleton instance to be published by the webserver
     */
    protected static PurityServlet  servletInstance;

    /**
     * mediator used to communicate with presences
     */
    protected BrillienMediator      mediator;


    static{
        servletInstance = new PurityServlet();
    }


    /**
     * Following the singleton model, retrieves the single instance created
     */
    public static PurityServlet getServletInstance() {
        return servletInstance;
    }



    {
        initMediatorContext();


    }

    protected void initMediatorContext() {

    }

    protected void setupMediatorContext(ServletConfig servletConfig) {

    }

    protected BrillienMediator getMediator() throws ServletException {
        if( mediator == null ){
            String username = getServletConfig().getInitParameter("username");
            String password = getServletConfig().getInitParameter("password");
            if( username == null ){
                username = "dispatcher" + this.getClass().getSimpleName(); password = "dispatcher";
            }
            try {
                if( mediator == null ){
                    setupMediatorContext( getServletConfig() );
                    mediator = BrillienContext.newMediator(username, password, new LoaderComplyer() {
                        @Override
                        public ClassLoader getLoader(String reference) {
                            try {
                                return BrillienServices.getPresenceManagerByName(
                                        BrillienServices.liaison, reference).getClassLoader();
                            } catch (BrillienException e) {
                                e.printStackTrace();
                                return this.getClass().getClassLoader();
                            }
                        }
                    });
                    mediator.connect();
                }
            } catch (BrillienException e) {
                throw new ServletException( e.getMessage(), e );
            }
        }
        return mediator;
    }

    @Override
    protected boolean checkApiKeyRequired(){
        boolean apiKeyRequired = super.checkApiKeyRequired();

        return  BrillienServices.liaison != null
                ? BrillienServices.liaison.getApiKeyPresenceName() != null
                : apiKeyRequired;
    }

    @Override
    protected boolean checkSsoRequired(){
        boolean ssoRequired = super.checkSsoRequired();

        return  BrillienServices.liaison != null
                ? BrillienServices.liaison.getSSOPresenceName() != null
                : ssoRequired;
    }

    @Override
    protected ApiKeyValidator getApiKeyValidator() throws ServletException {
        try {
            return (ApiKeyValidator)BrillienServices.getPresence( BrillienServices.liaison, BrillienServices.liaison.getApiKeyPresenceName() );
        } catch (BrillienException e) {
            throw new ServletException( e );
        }
    }

    @Override
    protected SessionValidator getSessionValidator() throws ServletException {
        try {
            return (SessionValidator)BrillienServices.getPresence( BrillienServices.liaison, BrillienServices.liaison.getSSOPresenceName() );
        } catch (BrillienException e) {
            throw new ServletException( e );
        }
    }

    @Override
    public void init() throws ServletException {
        super.init();

        getMediator();
    }


    @Override
    protected void sendError(HttpServletRequest req, HttpServletResponse resp, int errorCode, String errorMessage, Object errorValue) {
        try {
            //resp.setStatus( HttpStatus.SC_BAD_REQUEST );
            BrillienCommunication comm = mediator.newErrorCommunication(errorCode, errorMessage, errorValue);
            if( logger != null )
                logger.error( "Sending error message:" + comm );
            HttpServices.sendJSONMessage(req, resp, comm );
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    protected void processAsyncCall( final BrillienCommunication comm ){
        ThreadServices.getStpe().execute(new Runnable() {
            @Override
            public void run() {
                String originalSender = comm.getSender();
                comm.setSender( mediator.mediatorEntity() );
                try {
                    Object response = mediator.sendCommunication( comm );
                    try {
                        HttpServices.sendMessage(comm.getRedirectEntityName(), JsonServices.toJSON( response ));
                    } catch (StreamLineException e) {
                        e.printStackTrace();
                    }
                } catch (BrillienException e) {
                    JsonRpcError error = new JsonRpcError( e.getErrorCode(), e.getMessage(), e.getValue() );
                    if( logger != null )
                        logger.debug( "Sending async error message:" + error );
                    try {
                        HttpServices.sendMessage(comm.getRedirectEntityName(), JsonServices.toJSON(error));
                    } catch (StreamLineException e1) {
                        e1.printStackTrace();
                    }
                }
            }
        });
    }

    @Override
    protected Object call( String requestString, HashMap<String, String> clientData ) throws StreamLineException{
        final BrillienCommunication comm;
        try{
            comm = JsonServices.parseJSON( requestString, BrillienCommunication.class );
        } catch (SvensonRuntimeException sre){
            throw new StreamLineException( StreamLineException.ERROR_INVALID_PARAMS, "Message does not PureJSON-conform.", requestString );
        }

        comm.addExtension(RequestStatics.USER_DATA, clientData );

        if( comm.getMessageType() == Communication.TYPE_AGET ){
            processAsyncCall( comm );
            return "OK";
        }
        else
            try {
                String originalSender = comm.getSender();
                comm.setSender( mediator.mediatorEntity() );
                Object response = mediator.sendCommunication( comm );
                return response;
            } catch (BrillienException e) {
                throw new StreamLineException( e.getErrorCode(), e.getMessage(), e.getValue() );
            }
    }

    protected boolean checkCallPreconditions(HttpServletRequest req, HttpServletResponse resp){
        return true;
    }

}
