/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.services.websocket.applications;

import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.services.websocket.BrillienMediatorWebSocketApplication;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.json.JsonServices;
import org.glassfish.grizzly.websockets.WebSocket;

import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;

/**
 * Specialized websocket application to catch log records published to listening clients
 */
public class LogCatcherWebSocketApplication extends BrillienMediatorWebSocketApplication {

    /**
     * Singleton instance to be published by the webserver
     */
    protected static LogCatcherWebSocketApplication     webSocketApplicationInstance;

    protected static Level                              interstedLevel;

    static {
        init();

        interstedLevel = Level.INFO;
    }

    private static void init(){
        if( webSocketApplicationInstance == null )
            webSocketApplicationInstance = new LogCatcherWebSocketApplication();
    }

    /**
     * Following the singleton model, retrieves the single instance created
     */
    public static LogCatcherWebSocketApplication getWebSocketApplicationInstance() {
        init();

        return webSocketApplicationInstance;
    }

    @Override
    public void initWebSocketApplication() throws BrillienException {
        super.initWebSocketApplication();

        logger.info( this.getClass().getSimpleName() + " just joined to the log handlers...");

        BrillienContext.mainLogger.addHandler( new Handler() {
            @Override
            public synchronized Level getLevel() {
                return interstedLevel;
            }

            @Override
            public void publish(LogRecord logRecord) {
                try {
                    LogCatcherWebSocketApplication.this.publish( logRecord );
                } catch (StreamLineException e) { }
            }

            @Override
            public void flush() {
            }

            @Override
            public void close() throws SecurityException {
            }
        });
        BrillienContext.basicLog(Level.INFO, "LogCatcherWebSocketApplication is capturing logs...");
    }

    @Override
    public void onMessage(WebSocket socket, String text) {
        logger.info("Attempt to set log level:" + text);
        try{
            if( text != null && text.length()>0 )
                interstedLevel = Level.parse( text );
        } catch (Exception e){ logger.error( "Cannot convert to Level:" +  text, e ); }
    }

    public void publish(LogRecord record) throws StreamLineException {
        if( record.getLevel().intValue() >= interstedLevel.intValue() || record.getThrown() != null ){
            for (WebSocket webSocket : this.getWebSockets())
                if( webSocket.isConnected() ){
                    webSocket.send( JsonServices.toJSON(record) );
                }
        }
    }

}
